@@ -249,7 +249,9 @@ register_reply (DBusPendingCall *pending, void *user_data)
           dbus_message_iter_next (&iter_struct);
           dbus_message_iter_get_basic (&iter_struct, &obj_path);
 
+          g_free (app->desktop_name);
           app->desktop_name = g_strdup (app_name);
+          g_free (app->desktop_path);
           app->desktop_path = g_strdup (obj_path);
         }
     }
@@ -275,8 +277,8 @@ register_application (SpiBridge * app)
 
   /* These will be overridden when we get a reply, but in practice these
      defaults should always be correct */
-  app->desktop_name = ATSPI_DBUS_NAME_REGISTRY;
-  app->desktop_path = ATSPI_DBUS_PATH_ROOT;
+  app->desktop_name = g_strdup (ATSPI_DBUS_NAME_REGISTRY);
+  app->desktop_path = g_strdup (ATSPI_DBUS_PATH_ROOT);
 
   message = dbus_message_new_method_call (SPI_DBUS_NAME_REGISTRY,
                                           ATSPI_DBUS_PATH_ROOT,
@@ -289,6 +291,10 @@ register_application (SpiBridge * app)
     if (!dbus_connection_send_with_reply (app->bus, message, &pending, -1)
         || !pending)
     {
+        if (pending)
+          dbus_pending_call_unref (pending);
+
+        dbus_message_unref (message);
         return FALSE;
     }
 
@@ -339,6 +345,11 @@ deregister_application (SpiBridge * app)
     g_free (app->app_tmp_dir);
     app->app_tmp_dir = NULL;
   }
+
+  g_free (app->desktop_name);
+  app->desktop_name = NULL;
+  g_free (app->desktop_path);
+  app->desktop_path = NULL;
 }
 
 /*---------------------------------------------------------------------------*/
@@ -752,6 +763,27 @@ check_envvar (void)
     return TRUE;
 }
 
+void
+spi_atk_activate ()
+{
+  DRoutePath *treepath;
+
+  spi_atk_register_event_listeners ();
+  if (!spi_global_cache)
+    {
+      spi_global_cache    = g_object_new (SPI_CACHE_TYPE, NULL);
+      treepath = droute_add_one (spi_global_app_data->droute,
+                                 "/org/a11y/atspi/cache", spi_global_cache);
+
+      if (!treepath)
+        {
+          g_warning ("atk-bridge: Error in droute_add_one().  Already running?");
+          return;
+        }
+      spi_initialize_cache (treepath);
+    }
+}
+
 /*
  * spi_app_init
  *
@@ -771,7 +803,7 @@ atk_bridge_adaptor_init (gint * argc, gchar ** argv[])
   DBusError error;
   AtkObject *root;
   gboolean load_bridge;
-  DRoutePath *treepath, *accpath;
+  DRoutePath *accpath;
 
   load_bridge = check_envvar ();
   if (inited && !load_bridge)
@@ -845,21 +877,11 @@ atk_bridge_adaptor_init (gint * argc, gchar ** argv[])
    */
   spi_global_register = g_object_new (SPI_REGISTER_TYPE, NULL);
   spi_global_leasing  = g_object_new (SPI_LEASING_TYPE, NULL);
-  spi_global_cache    = g_object_new (SPI_CACHE_TYPE, NULL);
 
   /* Register droute for routing AT-SPI messages */
   spi_global_app_data->droute =
     droute_new ();
 
-  treepath = droute_add_one (spi_global_app_data->droute,
-                             "/org/a11y/atspi/cache", spi_global_cache);
-
-  if (!treepath)
-    {
-      g_warning ("atk-bridge: Error in droute_add_one().  Already running?");
-      return -1;
-    }
-
   accpath = droute_add_many (spi_global_app_data->droute,
                              "/org/a11y/atspi/accessible",
                              NULL,
@@ -870,7 +892,6 @@ atk_bridge_adaptor_init (gint * argc, gchar ** argv[])
 
 
   /* Register all interfaces with droute and set up application accessible db */
-  spi_initialize_cache (treepath);
   spi_initialize_accessible (accpath);
   spi_initialize_application (accpath);
   spi_initialize_action (accpath);
@@ -892,7 +913,7 @@ atk_bridge_adaptor_init (gint * argc, gchar ** argv[])
 
   /* Register methods to send D-Bus signals on certain ATK events */
   if (clients)
-    spi_atk_register_event_listeners ();
+    spi_atk_activate ();
 
   /* Set up filter and match rules to catch signals */
   dbus_bus_add_match (spi_global_app_data->bus, "type='signal', interface='org.a11y.atspi.Registry', sender='org.a11y.atspi.Registry'", NULL);
@@ -929,7 +950,9 @@ atk_bridge_adaptor_cleanup (void)
     {
       dbus_connection_remove_filter (spi_global_app_data->bus, signal_filter, NULL);
       droute_context_unregister (spi_global_app_data->droute, spi_global_app_data->bus);
+      dbus_connection_close (spi_global_app_data->bus);
       dbus_connection_unref (spi_global_app_data->bus);
+      spi_global_app_data->bus = NULL;
     }
 
   for (l = spi_global_app_data->direct_connections; l; l = l->next)
@@ -943,15 +966,16 @@ atk_bridge_adaptor_cleanup (void)
       dbus_connection_unref (connection);
     }
   g_list_free (spi_global_app_data->direct_connections);
+  spi_global_app_data->direct_connections = NULL;
 
   for (ls = clients; ls; ls = ls->next)
     g_free (l->data);
   g_slist_free (clients);
   clients = NULL;
 
-  g_object_unref (spi_global_cache);
-  g_object_unref (spi_global_leasing);
-  g_object_unref (spi_global_register);
+  g_clear_object (&spi_global_cache);
+  g_clear_object (&spi_global_leasing);
+  g_clear_object (&spi_global_register);
 
   if (spi_global_app_data->main_context)
     g_main_context_unref (spi_global_app_data->main_context);
@@ -981,7 +1005,7 @@ spi_atk_add_client (const char *bus_name)
       return;
   }
   if (!clients)
-    spi_atk_register_event_listeners ();
+    spi_atk_activate ();
   clients = g_slist_append (clients, g_strdup (bus_name));
   match = g_strdup_printf (name_match_tmpl, bus_name);
   dbus_bus_add_match (spi_global_app_data->bus, match, NULL);
